class SupamodelError(Exception):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)


class SupabaseAPIError(SupamodelError):
    def __init__(self, message: str, status_code: int, **kwargs):
        self.message = message
        self.status_code = status_code
        super().__init__(message, **kwargs)


class EmptyInputError(Exception):
    def __init__(
        self,
        data_input: list[dict] | dict,
        message: str = "The data entered shouldn't be empty. Either provide a list with a content of dictionaries or a dictionary with values within it.",
        **kwargs
    ):
        self.data_input = data_input
        self.message = message
        super().__init__(message, **kwargs)


class EmptyResponseError(Exception):
    def __init__(
        self,
        message: str = "The response from the database is empty. There is no data to process. We needed at least one record to proceed.",
        **kwargs
    ):
        self.message = message
        super().__init__(message, **kwargs)


class SupabaseConfigError(Exception):
    def __init__(
        self,
        message: str = "The Supabase `SUPABASE_URL` and `SUPABASE_KEY` is missing from environment variables, .env files, file secret, nor are they set. Please provide the Supabase URL and the Supabase Key.",
        **kwargs
    ):
        self.message = message
        super().__init__(message, **kwargs)



from datetime import datetime
from typing import Optional
from uuid import UUID

from pydantic import Field

from supamodel._abc import BaseModel
from supamodel.enums import OrderSide, OrderStatus, OrderType


class Trade(BaseModel):
    id: Optional[UUID] = Field(description="Unique identifier of the trade")
    position_id: UUID = Field(description="ID of the associated position")
    order_id: UUID = Field(description="ID of the associated order")
    quantity: float = Field(description="Quantity of the asset traded")
    price: float = Field(description="Price at which the trade occurred")
    fee: float = Field(description="Fee charged for the trade")
    timestamp: datetime = Field(description="Timestamp of when the trade occurred")
    created_at: Optional[datetime] = Field(
        description="Timestamp of when the trade record was created"
    )


class Order(BaseModel):
    id: Optional[UUID] = Field(description="Unique identifier of the order")
    portfolio_id: UUID = Field(description="ID of the associated portfolio")
    asset_id: UUID = Field(description="ID of the associated asset")
    exchange_id: UUID = Field(description="ID of the associated exchange")
    order_type: OrderType = Field(description="Type of the order (e.g., limit, market)")
    side: OrderSide = Field(description="Side of the order (buy or sell)")
    quantity: float = Field(description="Quantity of the asset to be traded")
    price: Optional[float] = Field(
        description="Price at which the order should be executed (for limit orders)"
    )
    status: OrderStatus = Field(description="Current status of the order")
    created_at: Optional[datetime] = Field(
        description="Timestamp of when the order was created"
    )
    updated_at: Optional[datetime] = Field(
        description="Timestamp of when the order was last updated"
    )



from pydantic import model_validator
from pydantic_settings import BaseSettings, SettingsConfigDict

from supamodel.exceptions import SupabaseConfigError


class Settings(BaseSettings):
    model_config = SettingsConfigDict(
        env_file=".env.dev", env_file_encoding="utf-8", extra="ignore"
    )


class SupabaseSettings(Settings):
    """
    Represents the settings for the Supabase client.

    Attributes:
        supabase_url (str | None): The URL of the Supabase server.
        supabase_key (str | None): The API key for accessing the Supabase server.
        schema (str): The default database schema.
        postgrest_client_timeout (int): The timeout value for the PostgREST client.
    """

    supabase_url: str | None = None
    supabase_key: str | None = None

    # -------------------------
    # client options
    # -------------------------
    schema: str = "public"
    postgrest_client_timeout: int = 30

    @model_validator(mode="after")
    @staticmethod
    def check_url_key_set(self):
        if not self.supabase_url or not self.supabase_key:
            raise SupabaseConfigError

    def options(self) -> dict:
        """
        Returns a dictionary of the Supabase client options.

        Returns:
            dict: A dictionary containing the Supabase client options.
        """
        return {
            "schema": self.schema,
            "postgrest_client_timeout": self.postgrest_client_timeout,
        }


settings = SupabaseSettings()



import abc
from typing import Optional, TypeVar
from uuid import UUID, uuid4

from pydantic import UUID4, ConfigDict, Field, computed_field

from supamodel._abc import BaseModel, IgnoreModel
from supamodel._types import CapitalStr
from supamodel.enums import Chain

DataT = TypeVar("DataT")


class ExchangeBase(BaseModel):
    name: str = Field(description="Name of the exchange")
    description: Optional[str] = None
    maker_fee: float = 0.0001
    taker_fee: float = 0.0001
    url: str | None = None


class Currency(IgnoreModel):
    code: str
    name: str


class BaseAsset(abc.ABC, BaseModel):
    id: UUID | None = None
    # name: str
    name: CapitalStr
    base: Optional[str] = Field(
        default=None,
        exclude=True,
        description="Base currency name used to find base currency ID. Hidden from model_dump",
        repr=False,
    )
    trade: Optional[str] = Field(
        default=None,
        exclude=True,
        description="Trade currency name used to find trade currency ID. Hidden from model_dump",
        repr=False,
    )
    exchange: Optional[CapitalStr] = Field(
        default=None,
        exclude=True,
        description="Exchange name used to find exchange ID. Hidden from model_dump",
        repr=False,
    )
    base_currency_id: Optional[UUID] = Field(None, description="Base currency ID")
    trade_currency_id: Optional[UUID] = Field(None, description="Trade currency ID")
    exchange_id: Optional[UUID] = Field(None, description="Exchange ID")

    # decimals: Optional[int] = Field(6, description="The number of integers to measure everything by.")
    # @computed_field
    @property
    def is_defi(self) -> bool:
        return False

    # @computed_field
    @property
    def is_wrapped(self) -> bool:
        return False


class ExchangeID(IgnoreModel):
    exchange_id: str
    name: str


class ChainDefaults(BaseModel):
    exchanges: Optional[list[CapitalStr]] = []
    currencies: Optional[list[Currency]] = []
    default_exchange: Optional[CapitalStr] = None
    default_currency: Optional[str] = None


class CurrencyID(Currency):
    currency_id: str

    @staticmethod
    def from_base(self, currency: Currency, currency_id: str) -> "CurrencyID":
        return CurrencyID(currency_id=currency_id, **currency.model_dump())


class ExchangeCreate(ExchangeBase):
    pass


class Exchange(ExchangeBase):
    id: UUID4 = Field(
        default_factory=uuid4,
        alias="exchange_id",
        description="Unique identifier of the exchange",
    )

    model_config = ConfigDict(from_attributes=True)


class Asset(BaseAsset):
    @computed_field
    def is_wrapped(self) -> bool:
        return False

    @computed_field
    def is_defi(self) -> bool:
        return False


class ChainAsset(BaseAsset):
    chain: Chain | None = None
    symbol: str | None = None
    decimals: Optional[int] = Field(
        6, description="The number of integers to measure everything by."
    )
    address: Optional[str] = Field(
        description="If the asset is a defi token, it'll have an address representing mint location."
    )
    is_meme: Optional[bool] = Field(False, description="Is the asset a meme token?")

    @computed_field
    def is_wrapped(self) -> bool:
        lower_name = self.name.lower()
        return "wrapped" in lower_name or "wrap" in lower_name

    @computed_field
    def is_defi(self) -> bool:
        return bool(self.address) and bool(self.chain)



from annotated_types import Annotated
from pydantic import BeforeValidator
from pydantic.functional_validators import AfterValidator
from typing_extensions import Annotated

from supamodel.utils import check_int, title_case

CapitalStr = Annotated[str, AfterValidator(title_case)]
ForceInt = Annotated[int, BeforeValidator(check_int)]



from supabase import create_client
from supabase.lib.client_options import ClientOptions

from supamodel.config import settings

client = create_client(
    supabase_url=settings.supabase_url,
    supabase_key=settings.supabase_key,
    options=ClientOptions(postgrest_client_timeout=60),
)



"""
Utility functions and classes.

This module contains utility functions and classes that are used throughout the package. Many are general-purpose and can be used in other projects as well. If they prove their worth, they may be moved to a separate package in the future.

The kinds of utilities found in this module include:

1. Custom data structures
2. Decorators
3. File I/O utilities
4. General-purpose functions
5. Logging utilities
6. Math utilities
7. String utilities
8. System utilities
9. Type utilities
10. Async utilities

"""

import asyncio
import functools
import inspect
import re
import zlib
from concurrent.futures import ThreadPoolExecutor
from functools import cached_property, wraps
from pathlib import Path
from typing import Any, Callable, Coroutine, TypeVar, cast

import diskcache
import orjson as json
import sniffio
import typer
from diskcache import UNKNOWN, Cache as LocalCache, memoize_stampede
from inflector import Inflector
from pydantic import BaseModel
from sniffio import current_async_library

# ---------------------------------------------------------
# Cache utilities
# ---------------------------------------------------------


class JSONDisk(diskcache.Disk):
    def __init__(self, directory, compress_level=1, **kwargs):
        self.compress_level = compress_level
        super().__init__(directory, **kwargs)

    def put(self, key):
        json_bytes = json.dumps(
            key,
            option=json.OPT_SERIALIZE_UUID
            | json.OPT_SORT_KEYS
            | json.OPT_NON_STR_KEYS
            | json.OPT_SERIALIZE_DATACLASS,
        )
        data = zlib.compress(json_bytes, self.compress_level)
        return super().put(data)

    def get(self, key, raw):
        data = super().get(key, raw)
        return json.loads(zlib.decompress(data).decode("utf-8"))

    def store(self, value, read, key=UNKNOWN):
        if not read:
            json_bytes = json.dumps(
                value,
                option=json.OPT_SERIALIZE_UUID
                | json.OPT_SORT_KEYS
                | json.OPT_NON_STR_KEYS
                | json.OPT_SERIALIZE_DATACLASS,
            )
            value = zlib.compress(json_bytes, self.compress_level)
        return super().store(value, read, key=key)

    def fetch(self, mode, filename, value, read):
        data = super().fetch(mode, filename, value, read)
        if not read:
            data = json.loads(zlib.decompress(data).decode("utf-8"))
        return data


class AppConfig(BaseModel):
    APP_NAME: str = "galapy"
    HOME: Path = Path.home()
    APP_DIR: str = typer.get_app_dir(APP_NAME, force_posix=True)

    @cached_property
    def app_dir(self) -> Path:
        return Path(self.APP_DIR)

    @cached_property
    def cache_dir(self) -> Path:
        return self.app_dir / "cache"

    @cached_property
    def config_dir(self) -> Path:
        return self.app_dir / "config"


app_config = AppConfig()

Cache = LocalCache(app_config.cache_dir, disk=JSONDisk)
memoize_stampede

# ---------------------------------------------------------
# Decorators
# ---------------------------------------------------------


class classproperty:
    """A decorator that allows a method to be accessed as a class property."""

    def __init__(self, method=None):
        self.fget = method

    def __get__(self, instance, cls=None):
        return self.fget(cls)

    def getter(self, method):
        self.fget = method
        return self


class cached_classattr(classproperty):
    """
    A decorator that defines a class-level property with caching mechanism.

    This decorator can be used to define a class-level property that is computed
    once and then cached for subsequent access. It is similar to the `@property`
    decorator, but it works at the class level instead of the instance level.

    Usage:
        class MyClass:
            @cached_classproperty
            def my_property(cls):
                # Compute and return the value of the property

    Note:
        This decorator is a subclass of the `classproperty` decorator, which is
        a decorator that allows defining class-level properties in Python.

    Attributes:
        fget (function): The method that computes the value of the property.

    Methods:
        get_result_field_name(): Returns the name of the result field used for caching.
        __get__(instance, cls=None): Retrieves the value of the property.
        getter(method): Sets the method that computes the value of the property.

    """

    def __init__(self, method=None):
        self.fget = method

    def get_result_field_name(self):
        return self.fget.__name__ + "_property_result" if self.fget else None

    def __get__(self, instance, cls=None):
        result_field_name = self.get_result_field_name()

        if hasattr(cls, result_field_name):
            return getattr(cls, result_field_name)

        if not cls or not result_field_name:
            return self.fget(cls)

        setattr(cls, result_field_name, self.fget(cls))
        return getattr(cls, result_field_name)

    def getter(self, method):
        self.fget = method
        return self


# ---------------------------------------------------------
# Async utilities
# ---------------------------------------------------------


def is_async_env() -> bool:
    """
    Check if a function is an asynchronous function.

    Args:
        func (Callable): The function to check.

    Returns:
        bool: True if the function is asynchronous, otherwise False.
    """
    try:
        current_async_library()

    except sniffio._impl.AsyncLibraryNotFoundError:
        return False

    return True


def expose_dual_runtimes(func):
    """
    Decorator that allows a function to be executed either synchronously or asynchronously,
    depending on whether an event loop is running or not.

    If the decorated function is a coroutine function, it checks if an event loop is running.
    If an event loop is running, the function is called directly. Otherwise, it is run synchronously
    using the `asyncio.run()` function.

    If the decorated function is not a coroutine, it is called directly.

    Args:
        func: The function to be decorated.

    Returns:
        The decorated function.

    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        if inspect.iscoroutinefunction(func):
            is_async = is_async_env()
            if not is_async:
                return run_sync(func(*args, **kwargs))
            # If an event loop is running, call the function directly
            return func(*args, **kwargs)
        else:
            # If the decorated function is not a coroutine, call it directly
            return func(*args, **kwargs)

    return wrapper


"""Utilities for working with asyncio."""


T = TypeVar("T")

BACKGROUND_TASKS = set()


def create_task(coro):
    """
    Creates async background tasks in a way that is safe from garbage
    collection.

    See
    https://textual.textualize.io/blog/2023/02/11/the-heisenbug-lurking-in-your-async-code/

    Example:

    async def my_coro(x: int) -> int:
        return x + 1

    # safely submits my_coro for background execution
    create_task(my_coro(1))
    """  # noqa: E501
    task = asyncio.create_task(coro)
    BACKGROUND_TASKS.add(task)
    task.add_done_callback(BACKGROUND_TASKS.discard)
    return task


async def run_async(fn: Callable[..., T], *args: Any, **kwargs: Any) -> T:
    """
    Runs a synchronous function in an asynchronous manner.

    Args:
        fn: The function to run.
        *args: Positional arguments to pass to the function.
        **kwargs: Keyword arguments to pass to the function.

    Returns:
        The return value of the function.

    Example:
        Basic usage:
        ```python
        def my_sync_function(x: int) -> int:
            return x + 1

        await run_async(my_sync_function, 1)
        ```
    """

    async def wrapper() -> T:
        try:
            return await loop.run_in_executor(
                None, functools.partial(fn, *args, **kwargs)
            )
        except Exception:
            # propagate the exception to the caller
            raise

    loop = asyncio.get_event_loop()
    return await wrapper()


def run_sync(coroutine: Coroutine[Any, Any, T]) -> T:
    """
    Runs a coroutine from a synchronous context, either in the current event
    loop or in a new one if there is no event loop running. The coroutine will
    block until it is done. A thread will be spawned to run the event loop if
    necessary, which allows coroutines to run in environments like Jupyter
    notebooks where the event loop runs on the main thread.

    Args:
        coroutine: The coroutine to run.

    Returns:
        The return value of the coroutine.

    Example:
        Basic usage:
        ```python
        async def my_async_function(x: int) -> int:
            return x + 1

        run_sync(my_async_function(1))
        ```
    """
    try:
        loop = asyncio.get_running_loop()
        if loop.is_running():
            with ThreadPoolExecutor() as executor:
                future = executor.submit(asyncio.run, coroutine)
                return future.result()
        else:
            return asyncio.run(coroutine)
    except RuntimeError:
        return asyncio.run(coroutine)


def run_sync_if_awaitable(obj: Any) -> Any:
    """
    If the object is awaitable, run it synchronously. Otherwise, return the
    object.

    Args:
        obj: The object to run.

    Returns:
        The return value of the object if it is awaitable, otherwise the object
        itself.

    Example:
        Basic usage:
        ```python
        async def my_async_function(x: int) -> int:
            return x + 1

        run_sync_if_awaitable(my_async_function(1))
        ```
    """
    return run_sync(obj) if inspect.isawaitable(obj) else obj


def make_sync(async_func):
    """
    Creates a synchronous function from an asynchronous function.
    """

    @functools.wraps(async_func)
    def sync_func(*args, **kwargs):
        return run_sync(async_func(*args, **kwargs))

    sync_func.__signature__ = inspect.signature(async_func)
    sync_func.__doc__ = async_func.__doc__
    return sync_func


class ExposeSyncMethodsMixin:
    """
    A mixin that can take functions decorated with `expose_sync_method`
    and automatically create synchronous versions.
    """

    def __init_subclass__(cls, **kwargs: Any) -> None:
        super().__init_subclass__(**kwargs)
        for method in list(cls.__dict__.values()):
            if callable(method) and hasattr(method, "_sync_name"):
                sync_method_name = method._sync_name
                setattr(cls, sync_method_name, method._sync_wrapper)


def expose_sync_method(name: str) -> Callable[..., Any]:
    """
    Decorator that automatically exposes synchronous versions of async methods.
    Note it doesn't work with classmethods.

    Args:
        name: The name of the synchronous method.

    Returns:
        The decorated function.

    Example:
        Basic usage:
        ```python
        class MyClass(ExposeSyncMethodsMixin):

            @expose_sync_method("my_method")
            async def my_method_async(self):
                return 42

        my_instance = MyClass()
        await my_instance.my_method_async() # returns 42
        my_instance.my_method()  # returns 42
        ```
    """

    def decorator(
        async_method: Callable[..., Coroutine[Any, Any, T]],
    ) -> Callable[..., Coroutine[Any, Any, T]]:
        @functools.wraps(async_method)
        def sync_wrapper(*args: Any, **kwargs: Any) -> T:
            coro = async_method(*args, **kwargs)
            return run_sync(coro)

        # Cast the sync_wrapper to the same type as the async_method to give the
        # type checker the needed information.
        casted_sync_wrapper = cast(Callable[..., T], sync_wrapper)

        # Attach attributes to the async wrapper
        setattr(async_method, "_sync_wrapper", casted_sync_wrapper)
        setattr(async_method, "_sync_name", name)

        # return the original async method; the sync wrapper will be added to
        # the class by the init hook
        return async_method

    return decorator


# ---------------------------------------------------------
# OOP utilities
# ---------------------------------------------------------


class Person:
    def __init__(self, name, age, address):
        self.name = name
        self.age = age
        self.address = address

    @classmethod
    def builder(cls):
        return PersonBuilder(cls)


class PersonBuilder:
    def __init__(self, person_class):
        self.person_class = person_class
        self.name = None
        self.age = None
        self.address = None

    def set_name(self, name):
        self.name = name
        return self

    def set_age(self, age):
        self.age = age
        return self

    def set_address(self, address):
        self.address = address
        return self

    def build(self):
        return self.person_class(self.name, self.age, self.address)


# Example usage:
# person1 = Person.builder().set_name("Alice").set_age(25).set_address("123 Main St").build()
# person2 = Person.builder().set_name("Bob").set_age(30).build()


# ---------------------------------------------------------
# Supabase utilities
# ---------------------------------------------------------


# ---------------------------------------------------------
# General-purpose functions
# ---------------------------------------------------------


def title_case(v: str) -> str:
    """Convert the value to title case."""
    split = v.split(" ")

    return " ".join([s.capitalize() for s in split])


def check_int(v: int) -> int:
    """Check if the value is an int."""
    try:
        return int(v)
    except ValueError as exc:
        raise TypeError("value must be an int") from exc


inflector = Inflector()


def pluralize(word):
    return inflector.pluralize(word)


def singularize(word):
    return inflector.singularize(word)


def tableize(word: str) -> str:
    return inflector.tableize(word)


def slugify(work: str) -> str:
    # Replace non-alphanumeric characters with hyphens
    slug = re.sub(r"[^a-zA-Z0-9]", "-", work)
    # Insert hyphens between camelCase words and convert to lower case
    slug = re.sub(r"(?<!^)(?=[A-Z][a-z])", "-", slug).lower()
    # Replace multiple consecutive hyphens with a single hyphen
    slug = re.sub(r"-+", "-", slug)

    return slug


def extract_first_word(s: str) -> str:
    s = s.strip()
    # Split the string into words using uppercase letters, underscores, or hyphens as separators
    words = re.split(r"(?=[A-Z])|_|-", s)
    # Return the first word in lowercase form
    return words[0].lower()


__all__ = [
    "memoize_stampede",
    "Cache",
]



import abc
import datetime
import random
from typing import Any, Dict, Generic, Optional, Set, TypeVar
from uuid import UUID

import pendulum
from pydantic import (
    Field,
    ValidationInfo,
    computed_field,
    field_validator,
    model_validator,
)

from supamodel._abc import BaseModel, Data

DataT = TypeVar("DataT", bound=Data)
InputModelT = TypeVar("InputModelT", bound=BaseModel)


class DataBox(BaseModel, Generic[DataT]):
    data: Optional[DataT] = None

    def is_list(self) -> bool:
        return isinstance(self.data, list)

    def is_dict(self) -> bool:
        return isinstance(self.data, dict)

    def pick_one(self) -> DataT:
        """
        Selects and returns a random item from the data.

        If the data is a list, a random item from the list is returned.
        If the data is a dictionary, a random item from the dictionary values is returned.
        If the data is neither a list nor a dictionary, the data itself is returned.

        Returns:
            Any: A random item from the data.
        """
        if self.is_list():
            return random.choice(self.data)
        elif self.is_dict():
            return random.choice(list(self.data.values()))
        return self.data

    # Add iterator methods if the data is a list
    def __iter__(self):
        if self.is_list():
            return iter(self.data)
        return iter([])

    def __getitem__(self, item):
        if self.is_list():
            return self.data[item]
        return None

    def __len__(self):
        if self.is_list():
            return len(self.data)
        return 0

    def __contains__(self, item):
        if self.is_list():
            return item in self.data
        return False


class AssetData(BaseModel, Generic[InputModelT]):
    """
    Represents asset data.

    Attributes:
        _asset_id (Union[UUID, str, None]): The asset ID.
        data (Union[list[InputModelT], InputModelT, None]): The asset data.

    Methods:
        asset_id (property): Gets the asset ID.
        asset_id (setter): Sets the asset ID.
        exclude_list: Excludes the first item from the data list.
        check_data: Checks the validity of the asset data.
        is_data_object: Checks if the asset data is an object.
        is_data_list: Checks if the asset data is a list.
        save_dump: Saves the asset data as a dictionary.
        asset_dump: Returns a dictionary representation of the asset data.
    """

    asset_id: UUID | str | None = Field(None, exclude=True, repr=False)
    data: list[InputModelT] | InputModelT | None = None

    # @computed_field
    # @property
    # def asset_id(self) -> str:
    #     return str(self._asset_id)

    # @asset_id.setter
    # def asset_id(self, value: str) -> str:
    #     self._asset_id = value

    def exclude_list(self) -> Dict:
        """
        Excludes the first item from the data list.

        Returns:
            None: If the data list is empty or not a list.
            Dict: The dictionary representation of the excluded item.
        """
        if self.is_list() and self.data:
            item = self.data[0]
            item.model_dump()
            return
        return

    @model_validator(mode="after")
    def check_data(self) -> Any:
        """
        Checks the validity of the asset data.

        Returns:
            Any: The instance of the class if the data is valid.

        Raises:
            ValueError: If the asset ID is None when adding data.
        """
        if self.data is None:
            return self
        if self.asset_id is None:
            raise ValueError("Asset ID is required when adding data")
        if isinstance(self.data, list):
            return self
        if isinstance(self.data, (BaseModel, Data)):
            self.data = [self.data]
            return self
        return self

    def is_data_object(self) -> bool:
        """
        Checks if the asset data is an object.

        Returns:
            bool: True if the asset data is an object, False otherwise.
        """
        return isinstance(self.data, InputModelT)

    def is_data_list(self) -> bool:
        """
        Checks if the asset data is a list.

        Returns:
            bool: True if the asset data is a list, False otherwise.
        """
        return isinstance(self.data, list)

    def save_dump(self) -> list[Dict]:
        """
        Saves the asset data as a dictionary.

        Returns:
            Dict: The dictionary representation of the asset data.
        """
        return self.model_dump(mode="json", exclude_none=True).get("data", [])

    def container_dump(self) -> list[Dict]:
        """
        Returns a dictionary representation of the asset data.

        Returns:
            Dict: The dictionary representation of the asset data.
        """
        records = self.save_dump()
        for record in records:
            record["asset_id"] = self.asset_id

        return records


class Handler(abc.ABC, BaseModel):

    @abc.abstractmethod
    def required(self) -> Set[str]:
        # Can use a set or list to store the required fields and exclude them from the model_dump
        # It's also possible to access those fields and add them into any list of dictionaries from a model dump.
        # It's an improvement to the AssetData class.
        raise NotImplementedError

    def required_dict(self) -> Dict:
        return self.model_dump(include=self.required(), exclude_none=True)

    def exclude_required(self) -> Dict:
        return self.model_dump(exclude=self.required(), exclude_none=True)


class Metric(BaseModel):
    type_: str = Field(
        ...,
        alias="type",
        description="The type of entity the metric is associated with (user, portfolio, performance, server, etc.).",
    )
    name: str = Field(..., alias="name", description="The name of the metric.")
    value: float = Field(..., alias="value", description="The value of the metric.")
    timestamp: datetime.datetime = Field(
        ...,
        alias="timestamp",
        description="The timestamp when the metric was recorded.",
    )
    metadata: Optional[Dict] = Field(
        {},
        alias="metadata",
        description="Additional metadata associated with the metric.",
    )


class TimeData(BaseModel):
    start_time: pendulum.DateTime = Field(
        ..., alias="startTime", description="The start time of the event or activity."
    )
    end_time: pendulum.DateTime = Field(
        ..., alias="endTime", description="The end time of the event or activity."
    )

    # duration: pendulum.Duration = Field(
    #     None,
    #     alias="duration",
    #     description="The duration of the event or activity (calculated from start_time and end_time)."
    # )
    timezone: str = Field(
        "UTC",
        alias="timezone",
        description="The timezone of the start_time and end_time.",
    )
    metadata: Optional[Dict] = Field(
        {},
        alias="metadata",
        description="Additional metadata associated with the time data.",
    )

    @computed_field
    def duration(self) -> pendulum.Duration:
        dur = self.end_time - self.start_time

        return dur

    @field_validator("end_time", mode="before")
    @classmethod
    def check_end_before_start(
        cls, val: pendulum.datetime, info: ValidationInfo
    ) -> pendulum.datetime:
        if (
            "start_time" in info.data
            and info.data.get("start_time")
            and val <= info.data.get("start_time")
        ):
            message = "end_time must be after start_time"
            raise ValueError(message)
        return val

    # @validator("end_time")
    # def end_time_after_start_time(cls, v, values):
    #     if "start_time" in values and v <= values["start_time"]:
    #         raise ValueError("end_time must be after start_time")
    #     return v


def main():
    data = Data(name="OpenBB", value=42)
    print(data)


if __name__ == "__main__":
    main()



import enum
from enum import Enum


class AssetClasses(str, enum.Enum):
    """
    Enumeration of asset classes.

    Attributes:
        EQUITY (str): Represents the equity asset class.
        BOND (str): Represents the bond asset class.
        CURRENCY (str): Represents the currency asset class.
        COMMODITY (str): Represents the commodity asset class.
        CRYPTO (str): Represents the cryptocurrency asset class.
        STOCK (str): Represents the stock asset class.
    """

    EQUITY = "EQUITY"
    BOND = "BOND"
    CURRENCY = "CURRENCY"
    COMMODITY = "COMMODITY"
    CRYPTO = "CRYPTO"
    STOCK = "STOCK"


class TimeInt(str, Enum):
    T_1m = "1m"
    T_3m = "3m"
    T_5m = "5m"
    T_15m = "15m"
    T_30m = "30m"
    T_1H = "1H"
    T_2H = "2H"
    T_4H = "4H"
    T_6H = "6H"
    T_8H = "8H"
    T_12H = "12H"
    T_1D = "1D"
    T_3D = "3D"
    T_1W = "1W"
    T_1M = "1M"


class Chain(str, Enum):
    SOLANA = "solana"
    TERRA = "ethereum"
    POLYGON = "polygon"
    AVALANCHE = "avalanche"
    ARBITRUM = "arbitrum"
    OPTIMISM = "optimism"


class RunState(str, Enum):
    RUNNING = "RUNNING"
    STOPPED = "STOPPED"
    PAUSED = "PAUSED"
    COMPLETED = "COMPLETED"
    FAILED = "FAILED"
    PENDING = "PENDING"


class SystemState(str, Enum):
    LIVE = "LIVE"
    BACKTEST = "BACKTEST"
    SIMULATION = "SIMULATION"
    PAPER = "PAPER"
    DEMO = "DEMO"
    DEVELOPMENT = "DEVELOPMENT"


class OrderStatus(str, Enum):
    PENDING = "pending"
    FILLED = "filled"
    PARTIALLY_FILLED = "partially_filled"
    CANCELED = "canceled"
    REJECTED = "rejected"


class OrderSide(str, Enum):
    BUY = "buy"
    SELL = "sell"


class OrderType(str, Enum):
    LIMIT = "limit"
    MARKET = "market"
    STOP = "stop"
    STOP_LIMIT = "stop_limit"
    TRAILING_STOP = "trailing_stop"
    TAKE_PROFIT = "take_profit"
    TAKE_PROFIT_LIMIT = "take_profit_limit"
    LIMIT_MAKER = "limit_maker"


class TimeIntSQL(str, Enum):
    T_1m = "1 minute"
    T_3m = "3 minutes"
    T_5m = "5 minutes"
    T_15m = "15 minutes"
    T_30m = "30 minutes"
    T_1H = "1 hour"
    T_2H = "2 hours"
    T_4H = "4 hours"
    T_6H = "6 hours"
    T_8H = "8 hours"
    T_12H = "12 hours"
    T_1D = "1 day"
    T_3D = "3 days"
    T_1W = "1 week"
    T_1M = "1 month"



from datetime import datetime
from typing import Any, Optional
from uuid import UUID

import pendulum as pdm
from pydantic import Field, model_validator

from supamodel._abc import BaseModel, CamelModel, Data
from supamodel.enums import TimeIntSQL


def now_utc():
    return pdm.now("UTC")


class PriceBar(CamelModel):
    address: str | None = Field(None, repr=False)
    interval: TimeIntSQL = Field(..., alias="type")
    close: float = Field(..., alias="c")
    high: float = Field(..., alias="h")
    low: float = Field(..., alias="l")
    open: float = Field(..., alias="o")
    volume: int = Field(..., alias="v")
    timestamp: datetime = Field(..., alias="unixTime")


class Price(Data):
    address: str = Field(..., exclude=True, repr=False)
    price: float
    timestamp: datetime


class PriceBarInput(BaseModel):
    asset_id: Optional[UUID] = None
    prices: list[PriceBar] | PriceBar | None = None

    @model_validator(mode="after")
    def convert_price_list(self) -> Any:
        if self.prices is None:
            return self

        if self.asset_id is None:
            raise ValueError("Asset ID is required")
        if isinstance(self.prices, list):
            return self
        if isinstance(self.prices, PriceBar):
            self.prices = [self.prices]

            return self
        return self
